Attribute VB_Name = "MMsgBoxPos"
' *************************************************************************
'  Copyright 2000 Karl E. Peterson
'  All Rights Reserved, http://www.mvps.org/vb
' *************************************************************************
'  You are free to use this code within your own applications, but you
'  are expressly forbidden from selling or otherwise distributing this
'  source code, non-compiled, without prior written consent.
' *************************************************************************
Option Explicit

' Win32 APIs
Private Declare Function SetWindowsHookEx Lib "user32" Alias "SetWindowsHookExA" (ByVal idHook As Long, ByVal lpfn As Long, ByVal hmod As Long, ByVal dwThreadId As Long) As Long
Private Declare Function UnhookWindowsHookEx Lib "user32" (ByVal hHook As Long) As Long
Private Declare Function GetWindowRect Lib "user32" (ByVal hWnd As Long, lpRect As RECT) As Long
Private Declare Function MoveWindow Lib "user32" (ByVal hWnd As Long, ByVal x As Long, ByVal y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal bRepaint As Long) As Long
Private Declare Function SetTimer Lib "user32" (ByVal hWnd As Long, ByVal nIDEvent As Long, ByVal uElapse As Long, ByVal lpTimerFunc As Long) As Long
Private Declare Function KillTimer Lib "user32" (ByVal hWnd As Long, ByVal nIDEvent As Long) As Long
Private Declare Function IsWindow Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function SendMessage Lib "user32" Alias "SendMessageA" (ByVal hWnd As Long, ByVal wMsg As Long, ByVal wParam As Long, lParam As Any) As Long
Private Declare Function PostMessage Lib "user32" Alias "PostMessageA" (ByVal hWnd As Long, ByVal wMsg As Long, ByVal wParam As Long, lParam As Any) As Long

Private Type RECT
   Left As Long
   Top As Long
   Right As Long
   Bottom As Long
End Type

' Dismissal message
Private Const WM_CLOSE = &H10

' SetWindowsHook() codes
Private Const WH_MIN = (-1)
Private Const WH_MSGFILTER = (-1)
Private Const WH_JOURNALRECORD = 0
Private Const WH_JOURNALPLAYBACK = 1
Private Const WH_KEYBOARD = 2
Private Const WH_GETMESSAGE = 3
Private Const WH_CALLWNDPROC = 4
Private Const WH_CBT = 5
Private Const WH_SYSMSGFILTER = 6
Private Const WH_MOUSE = 7
Private Const WH_HARDWARE = 8
Private Const WH_DEBUG = 9
Private Const WH_SHELL = 10
Private Const WH_FOREGROUNDIDLE = 11
Private Const WH_MAX = 11

' CBT Hook Codes
Private Const HCBT_MOVESIZE = 0
Private Const HCBT_MINMAX = 1
Private Const HCBT_QS = 2
Private Const HCBT_CREATEWND = 3
Private Const HCBT_DESTROYWND = 4
Private Const HCBT_ACTIVATE = 5
Private Const HCBT_CLICKSKIPPED = 6
Private Const HCBT_KEYSKIPPED = 7
Private Const HCBT_SYSCOMMAND = 8
Private Const HCBT_SETFOCUS = 9

' HCBT_ACTIVATE structure pointed to by lParam
Private Type CBTACTIVATESTRUCT
   fMouse As Long                 'BOOL
   hWndActive As Long
End Type

' Private module vars
Private m_hHook As Long
Private m_TmrID As Long
Private m_hWnd As Long
Private m_hWndOver As Long
Private m_Top As Long
Private m_Left As Long

' **********************************************
'  Public routines to display MsgBox
' **********************************************
Public Function MsgBoxAbs(ByVal Prompt As String, Optional ByVal Buttons As VbMsgBoxStyle = vbOKOnly, Optional Title As String, Optional ByVal Left As Long, Optional ByVal Top As Long, Optional ByVal Timeout As Long) As VbMsgBoxResult
   ' Cache top/left values.
   m_Top = Top
   m_Left = Left
   ' Set default title.
   If Title = "" Then Title = App.Title
   ' Set the CBT hook.
   m_hHook = SetWindowsHookEx(WH_CBT, AddressOf CBTProcAbs, App.hInstance, App.ThreadID)
   ' Start timer, if requested.
   If Timeout > 0 Then
      m_TmrID = SetTimer(0&, 0&, Timeout, AddressOf TimerProc)
   End If
   ' Display the message box.
   MsgBoxAbs = MsgBox(Prompt, Buttons, Title)
End Function

Public Function MsgBoxOver(ByVal Prompt As String, Optional ByVal Buttons As VbMsgBoxStyle = vbOKOnly, Optional Title As String, Optional ByVal hWndOver As Long, Optional ByVal Timeout As Long) As VbMsgBoxResult
   ' Cache hWndOver value.
   m_hWndOver = hWndOver
   ' Set default title.
   If Title = "" Then Title = App.Title
   ' Set the CBT hook.
   m_hHook = SetWindowsHookEx(WH_CBT, AddressOf CBTProcOver, App.hInstance, App.ThreadID)
   ' Start timer, if requested.
   If Timeout > 0 Then
      m_TmrID = SetTimer(0&, 0&, Timeout, AddressOf TimerProc)
   End If
   ' Display the message box.
   MsgBoxOver = MsgBox(Prompt, Buttons, Title)
End Function

' **********************************************
'  Private hook callback routines.
' **********************************************
Private Function CBTProcAbs(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
   Dim r As RECT
   ' We only want to react to activations.
   If nCode = HCBT_ACTIVATE Then
      ' Handle is in wParam.
      m_hWnd = wParam
      ' Get width/height of msgbox.
      Call GetWindowRect(m_hWnd, r)
      ' Adjust msgbox coordinates.
      r.Right = m_Left + (r.Right - r.Left)
      r.Bottom = m_Top + (r.Bottom - r.Top)
      r.Left = m_Left
      r.Top = m_Top
      ' Optionally, validate new coordinates.
      Call ValidateCoords(r)
      ' Move to new position.
      Call MoveWindow(m_hWnd, r.Left, r.Top, r.Right - r.Left, r.Bottom - r.Top, True)
      ' Hook no longer needed.
      Call UnhookWindowsHookEx(m_hHook)
   End If
End Function

Private Function CBTProcOver(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
   Dim r1 As RECT, r2 As RECT, r3 As RECT
   ' We only want to react to activations.
   If nCode = HCBT_ACTIVATE Then
      ' Handle is in wParam.
      m_hWnd = wParam
      ' Get width/height of msgbox.
      Call GetWindowRect(m_hWnd, r1)
      ' Get coordinates for owner window.
      Call GetWindowRect(m_hWndOver, r2)
      ' Adjust msgbox coordinates.
      r3.Left = r2.Left + ((r2.Right - r2.Left) \ 2) _
                - ((r1.Right - r1.Left) \ 2)
      r3.Right = r3.Left + (r1.Right - r1.Left)
      r3.Top = r2.Top + ((r2.Bottom - r2.Top) \ 2) _
               - ((r1.Bottom - r1.Top) \ 2)
      r3.Bottom = r3.Top + (r1.Bottom - r1.Top)
      ' Optionally, validate new coordinates.
      Call ValidateCoords(r3)
      ' Move to new position.
      Call MoveWindow(m_hWnd, r3.Left, r3.Top, r3.Right - r3.Left, r3.Bottom - r3.Top, True)
      ' Hook no longer needed.
      Call UnhookWindowsHookEx(m_hHook)
   End If
End Function

' **********************************************
'  Private timer callback
' **********************************************
Private Function TimerProc(ByVal hWnd As Long, ByVal uMsg As Long, ByVal idEvent As Long, ByVal dwTime As Long) As Long
   ' No default value is defined by Windows
   ' for vbAbortRetryIgnore and vbYesNo
   ' message box styles. We'll try posting a
   ' WM_CLOSE message anyway, but it will
   ' be useless if either of those are set.
   If IsWindow(m_hWnd) Then
      Call PostMessage(m_hWnd, WM_CLOSE, 0, ByVal 0&)
   End If
   ' Kill timer.
   If m_TmrID Then
      Call KillTimer(0&, m_TmrID)
      m_TmrID = 0
   End If
End Function

' **********************************************
'  Other private methods
' **********************************************
Private Sub ValidateCoords(r As RECT)
   ' Check that we're not off left edge.
   If r.Left < 0 Then
      r.Right = r.Right - r.Left
      r.Left = 0
   End If
   ' Check that we're not over the top.
   If r.Top < 0 Then
      r.Bottom = r.Bottom - r.Top
      r.Top = 0
   End If
   ' Check right/bottom edges.
   With Screen
      If (r.Right * .TwipsPerPixelX) > .Width Then
         r.Left = (.Width \ .TwipsPerPixelX) - (r.Right - r.Left)
         r.Right = (.Width \ .TwipsPerPixelX)
      End If
      If (r.Bottom * .TwipsPerPixelY) > .Height Then
         r.Top = (.Height \ .TwipsPerPixelY) - (r.Bottom - r.Top)
         r.Bottom = (.Height \ .TwipsPerPixelY)
      End If
   End With
End Sub
